# Generated by Django 2.2.5 on 2019-10-06 20:31

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('authentication', '0016_ownershiprecord'),
        ('eveonline', '0010_alliance_ticker'),
    ]

    operations = [
        migrations.CreateModel(
            name='Freight',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'default_permissions': (),
                'permissions': (('basic_access', 'Can access this app'), ('setup_contract_handler', 'Can setup contract handler'), ('use_calculator', 'Can use the calculator'), ('view_contracts', 'Can view the contracts list'), ('add_location', 'Can add / update locations'), ('view_statistics', 'Can view freight statistics')),
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='Location',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=100)),
                ('solar_system_id', models.IntegerField(blank=True, default=None, null=True)),
                ('type_id', models.IntegerField(blank=True, default=None, null=True)),
                ('category_id', models.IntegerField(choices=[(3, 'station'), (65, 'structure'), (0, '(unknown)')], default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Pricing',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('active', models.BooleanField(default=True, help_text='Non active pricings will not be used or shown')),
                ('price_base', models.FloatField(blank=True, default=0, help_text='Base price in ISK')),
                ('price_per_volume', models.FloatField(blank=True, default=0, help_text='Add-on price per m3 volume in ISK')),
                ('price_per_collateral_percent', models.FloatField(blank=True, default=0, help_text='Add-on price in % of collaterial')),
                ('collateral_min', models.BigIntegerField(blank=True, default=0, help_text='Minimum required collateral in ISK')),
                ('collateral_max', models.BigIntegerField(blank=True, default=None, help_text='Maximum allowed collateral in ISK', null=True)),
                ('volume_max', models.FloatField(blank=True, default=None, help_text='Maximum allowed volume in m3', null=True)),
                ('days_to_expire', models.IntegerField(blank=True, default=None, help_text='Recommended days for contracts to expire', null=True)),
                ('days_to_complete', models.IntegerField(blank=True, default=None, help_text='Recommended days for contract completion', null=True)),
                ('details', models.TextField(blank=True, default=None, help_text='Text with additional instructions for using this pricing', null=True)),
                ('end_location', models.ForeignKey(help_text='Destination station or structure for courier route', on_delete=django.db.models.deletion.CASCADE, related_name='pricing_end_location', to='freight.Location')),
                ('start_location', models.ForeignKey(help_text='Starting station or structure for courier route', on_delete=django.db.models.deletion.CASCADE, related_name='pricing_start_location', to='freight.Location')),
            ],
            options={
                'unique_together': {('start_location', 'end_location')},
            },
        ),
        migrations.CreateModel(
            name='ContractHandler',
            fields=[
                ('alliance', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='eveonline.EveAllianceInfo')),
                ('version_hash', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('last_sync', models.DateTimeField(blank=True, default=None, null=True)),
                ('character', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='authentication.CharacterOwnership')),
            ],
        ),
        migrations.CreateModel(
            name='Contract',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contract_id', models.IntegerField()),
                ('collateral', models.FloatField()),
                ('date_accepted', models.DateTimeField(default=None, null=True)),
                ('date_completed', models.DateTimeField(default=None, null=True)),
                ('date_expired', models.DateTimeField()),
                ('date_issued', models.DateTimeField()),
                ('days_to_complete', models.IntegerField()),
                ('for_corporation', models.BooleanField()),
                ('price', models.FloatField()),
                ('reward', models.FloatField()),
                ('status', models.CharField(choices=[('outstanding', 'outstanding'), ('in_progress', 'in progress'), ('finished_issuer', 'finished issuer'), ('finished_contractor', 'finished contractor'), ('finished', 'finished'), ('canceled', 'canceled'), ('rejected', 'rejected'), ('failed', 'failed'), ('deleted', 'deleted'), ('reversed', 'reversed')], max_length=32)),
                ('title', models.CharField(default=None, max_length=100, null=True)),
                ('volume', models.FloatField()),
                ('date_notified', models.DateTimeField(default=None, help_text='datetime of latest notification, None = none has been sent', null=True)),
                ('acceptor', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='contract_acceptor', to='eveonline.EveCharacter')),
                ('end_location', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contract_end_location', to='freight.Location')),
                ('handler', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='freight.ContractHandler')),
                ('issuer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contract_issuer', to='eveonline.EveCharacter')),
                ('issuer_corporation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contract_issuer', to='eveonline.EveCorporationInfo')),
                ('pricing', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='freight.Pricing')),
                ('start_location', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contract_start_location', to='freight.Location')),
            ],
        ),
        migrations.AddIndex(
            model_name='contract',
            index=models.Index(fields=['status'], name='freight_con_status_b4eb08_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='contract',
            unique_together={('handler', 'contract_id')},
        ),
        migrations.AlterField(
            model_name='location',
            name='id',
            field=models.BigIntegerField(primary_key=True, serialize=False, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='collateral_min',
            field=models.BigIntegerField(blank=True, default=None, help_text='Minimum required collateral in ISK', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='days_to_complete',
            field=models.IntegerField(blank=True, default=None, help_text='Recommended days for contract completion', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='days_to_expire',
            field=models.IntegerField(blank=True, default=None, help_text='Recommended days for contracts to expire', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='price_base',
            field=models.FloatField(blank=True, default=None, help_text='Base price in ISK', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AddField(
            model_name='pricing',
            name='price_min',
            field=models.FloatField(blank=True, default=None, help_text='Minimum total price in ISK', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='price_per_collateral_percent',
            field=models.FloatField(blank=True, default=None, help_text='Add-on price in % of collaterial', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='price_per_volume',
            field=models.FloatField(blank=True, default=None, help_text='Add-on price per m3 volume in ISK', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='collateral_max',
            field=models.BigIntegerField(blank=True, default=None, help_text='Maximum allowed collateral in ISK', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='days_to_complete',
            field=models.IntegerField(blank=True, default=None, help_text='Recommended days for contract completion', null=True, validators=[django.core.validators.MinValueValidator(1)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='days_to_expire',
            field=models.IntegerField(blank=True, default=None, help_text='Recommended days for contracts to expire', null=True, validators=[django.core.validators.MinValueValidator(1)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='volume_max',
            field=models.FloatField(blank=True, default=None, help_text='Maximum allowed volume in m3', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AddField(
            model_name='pricing',
            name='volume_min',
            field=models.FloatField(blank=True, default=None, help_text='Minimum allowed volume in m3', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterModelOptions(
            name='freight',
            options={'default_permissions': (), 'managed': False, 'permissions': (('add_location', 'Can add / update locations'), ('basic_access', 'Can access this app'), ('setup_contract_handler', 'Can setup contract handler'), ('use_calculator', 'Can use the calculator'), ('view_contracts', 'Can view the contracts list'), ('view_statistics', 'Can view freight statistics'))},
        ),
        migrations.RemoveField(
            model_name='contract',
            name='date_notified',
        ),
        migrations.AddField(
            model_name='contract',
            name='issues',
            field=models.TextField(default=None, help_text='List or price check issues as JSON array of strings or None', null=True),
        ),
        migrations.AddField(
            model_name='contracthandler',
            name='last_error',
            field=models.IntegerField(choices=[(0, 'No error'), (1, 'Invalid token'), (2, 'Expired token'), (3, 'Insufficient permissions'), (4, 'No character set for fetching alliance contacts'), (5, 'ESI API is currently unavailable'), (99, 'Unknown error')], default=0, help_text='error that occurred at the last sync atttempt (if any)'),
        ),
        migrations.AlterField(
            model_name='contracthandler',
            name='last_sync',
            field=models.DateTimeField(blank=True, default=None, help_text='when the last sync happened', null=True),
        ),
        migrations.AlterField(
            model_name='contracthandler',
            name='version_hash',
            field=models.CharField(blank=True, default=None, help_text='hash to identify changes to contracts', max_length=32, null=True),
        ),
        migrations.AlterField(
            model_name='location',
            name='category_id',
            field=models.IntegerField(choices=[(3, 'station'), (65, 'structure'), (0, '(unknown)')], default=0, help_text='Eve Online category ID', validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='location',
            name='id',
            field=models.BigIntegerField(help_text='Eve Online location ID, either item ID for stations or structure ID for structures', primary_key=True, serialize=False, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='location',
            name='name',
            field=models.CharField(help_text='In-game name of this station or structure', max_length=100),
        ),
        migrations.AlterField(
            model_name='location',
            name='solar_system_id',
            field=models.IntegerField(blank=True, default=None, help_text='Eve Online solar system ID', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AlterField(
            model_name='location',
            name='type_id',
            field=models.IntegerField(blank=True, default=None, help_text='Eve Online type ID', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
        migrations.AddField(
            model_name='contract',
            name='date_notified',
            field=models.DateTimeField(default=None, help_text='datetime of latest notification, None = none has been sent', null=True),
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-28 21:55

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    atomic = False

    dependencies = [
        ('freight', '0004_remove_contract_price'),
    ]

    operations = [
        migrations.CreateModel(
            name='EveOrganization',
            fields=[
                ('id', models.IntegerField(primary_key=True, serialize=False, validators=[django.core.validators.MinValueValidator(0)])),
                ('category', models.CharField(choices=[('alliance', 'Alliance'), ('corporation', 'Corporation')], max_length=32)),
                ('name', models.CharField(max_length=254)),
            ],
        ),
        migrations.RenameField(
            model_name='contracthandler',
            old_name='alliance',
            new_name='organization',
        ),
        migrations.AlterField(
            model_name='contract',
            name='handler',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='freight.ContractHandler', to_field='organization'),
        ),
        migrations.AlterField(
            model_name='contracthandler',
            name='character',
            field=models.ForeignKey(default=None, help_text='character used for syncing contracts', null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='authentication.CharacterOwnership'),
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-14 17:12

from django.db import migrations


class Migration(migrations.Migration):

    atomic = False
    
    dependencies = [
        ('freight', '0013_auto_20191214_1522'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='EveOrganization',
            new_name='EveEntity',
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-28 10:23

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0002_auto_20191015_1229'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contract',
            name='acceptor',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='contract_acceptor', to='eveonline.EveCharacter'),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_accepted',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_completed',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_notified',
            field=models.DateTimeField(blank=True, default=None, help_text='datetime of latest notification, None = none has been sent', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='issues',
            field=models.TextField(blank=True, default=None, help_text='List or price check issues as JSON array of strings or None', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='pricing',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='freight.Pricing'),
        ),
        migrations.AlterField(
            model_name='contract',
            name='title',
            field=models.CharField(blank=True, default=None, max_length=100, null=True),
        ),
    ]



import json
import logging
from time import sleep

from bravado.exception import HTTPUnauthorized, HTTPForbidden

from django.db import models, transaction

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.eveonline.providers import ObjectNotFound

from .app_settings import (
    FREIGHT_DISCORD_WEBHOOK_URL, 
    FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL,   
)
from .helpers import EsiSmartRequest
from .utils import LoggerAddTag, make_logger_prefix


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


class PricingManager(models.Manager):

    def get_default(self):
        """return the default pricing if defined
        else the first pricing, which can be None if no pricing exists
        """
        pricing_qs = self.filter(is_active=True)        
        pricing = pricing_qs.filter(is_default=True).first()
        if not pricing:        
            pricing = pricing_qs.first()
        
        return pricing

    def get_or_default(self, pk: int = None):
        """returns the pricing for given pk if found else default pricing"""
        if pk:
            try:
                pricing = self.filter(is_active=True).get(pk=pk)
            except self.model.DoesNotExist:
                pricing = self.get_default()
        else:            
            pricing = self.get_default()
        return pricing


class LocationManager(models.Manager):
    STATION_ID_START = 60000000
    STATION_ID_END = 69999999
    
    def get_or_create_from_esi(
        self, 
        esi_client: object, 
        location_id: int,
        add_unknown: bool = True
    ) -> tuple:
        """gets or creates location object with data fetched from ESI"""
        from .models import Location
        try:
            location = self.get(id=location_id)
            created = False
        except Location.DoesNotExist:
            location, created = self.update_or_create_from_esi(
                esi_client, location_id, add_unknown
            )
        
        return location, created

    def update_or_create_from_esi(
        self, 
        esi_client: object, 
        location_id: int, 
        add_unknown: bool = True
    ) -> tuple:
        """updates or creates location object with data fetched from ESI"""
        from .models import Location

        add_prefix = make_logger_prefix(location_id)

        if (
            location_id >= self.STATION_ID_START 
            and location_id <= self.STATION_ID_END
        ):
            logger.info(add_prefix('Fetching station from ESI'))
            try:                
                station = EsiSmartRequest.fetch(
                    'Universe.get_universe_stations_station_id',
                    args={'station_id': location_id},
                    esi_client=esi_client,
                    logger_tag=add_prefix()
                )
                location, created = self.update_or_create(
                    id=location_id,
                    defaults={
                        'name': station['name'],                    
                        'solar_system_id': station['system_id'],
                        'type_id': station['type_id'],                    
                        'category_id': Location.CATEGORY_STATION_ID
                    }
                ) 
            except Exception as ex:
                logger.exception(add_prefix(
                    'Failed to load station: {}'.format(ex)
                ))
                raise ex
            
        else:            
            try:                
                structure = EsiSmartRequest.fetch(
                    'Universe.get_universe_structures_structure_id',
                    args={'structure_id': location_id},
                    esi_client=esi_client,
                    logger_tag=add_prefix()
                )
                location, created = self.update_or_create(
                    id=location_id,
                    defaults={
                        'name': structure['name'],                    
                        'solar_system_id': structure['solar_system_id'],
                        'type_id': structure['type_id'],
                        'category_id': Location.CATEGORY_STRUCTURE_ID
                    }
                )      
            except (HTTPUnauthorized, HTTPForbidden) as ex:
                logger.warn(add_prefix(
                    'No access to this structure: {}'.format(ex)
                ))      
                if add_unknown:
                    location, created = self.get_or_create(
                        id=location_id,
                        defaults={
                            'name': 'Unknown structure {}'.format(location_id),
                            'category_id': Location.CATEGORY_STRUCTURE_ID
                        }
                    )
                else:
                    raise ex
            except Exception as ex:
                logger.exception(add_prefix(
                    'Failed to load structure: {}'.format(ex)
                ))      
                raise ex
       
        return location, created


class EveEntityManager(models.Manager):
    
    def get_or_create_from_esi(self, id: int) -> tuple:
        """gets or creates entity object with data fetched from ESI"""
        from .models import EveEntity
        try:
            entity = self.get(id=id)
            created = False
        except EveEntity.DoesNotExist:
            entity, created = self.update_or_create_from_esi(id)
        
        return entity, created

    def update_or_create_from_esi(self, id: int) -> tuple:
        """updates or creates entity object with data fetched from ESI"""
        
        add_prefix = make_logger_prefix(id)
                
        try:            
            response = EsiSmartRequest.fetch(
                esi_path='Universe.post_universe_names', 
                args={'ids': [id]},
                logger_tag=add_prefix()
            )
            if len(response) != 1:
                raise ObjectNotFound(id, 'unknown_type')
            else:
                entity_data = response[0]
            entity, created = self.update_or_create(
                id=entity_data['id'],
                defaults={
                    'name': entity_data['name'],
                    'category': entity_data['category'],
                }
            ) 
        except Exception as ex:
            logger.exception(add_prefix(
                'Failed to load entity with id {} from ESI: {}'.format(id, ex)
            ))
            raise ex
        
        return entity, created

    def update_or_create_from_evecharacter(
        self,             
        character: EveCharacter,
        category: str
    ) -> tuple:
        """updates or creates EveEntity object from an EveCharacter object"""
        from .models import EveEntity
        
        add_prefix = make_logger_prefix(character.character_id)

        try:            
            if category == EveEntity.CATEGORY_ALLIANCE:            
                if not character.alliance_id:
                    raise ValueError('character is not an alliance member')
                eve_entity, created = self.update_or_create(
                    id=character.alliance_id,
                    defaults={
                        'name': character.alliance_name,
                        'category': EveEntity.CATEGORY_ALLIANCE,
                    }
                )
            elif category == EveEntity.CATEGORY_CORPORATION:
                eve_entity, created = self.update_or_create(
                    id=character.corporation_id,
                    defaults={
                        'name': character.corporation_name,
                        'category': EveEntity.CATEGORY_CORPORATION,
                    }
                )
            elif category == EveEntity.CATEGORY_CHARACTER:
                eve_entity, created = self.update_or_create(
                    id=character.character_id,
                    defaults={
                        'name': character.character_name,
                        'category': EveEntity.CATEGORY_CHARACTER,
                    }
                )
            else:
                raise ValueError('Invalid category: {}'. format(category))
        
        except Exception as ex:
            logger.exception(add_prefix(
                'Failed to convert to EveEntity: {}'.format(ex)
            ))
            raise ex

        return eve_entity, created


class ContractManager(models.Manager):
    
    def update_or_create_from_dict(
        self, handler: object, contract: dict, esi_client: object
    ) -> tuple:
        """updates or creates a contract from given dict"""        
        acceptor, acceptor_corporation = \
            self._identify_contract_acceptor(contract)
        issuer_corporation, issuer = \
            self._identify_contract_issuer(contract)        
        date_accepted = contract['date_accepted'] \
            if 'date_accepted' in contract else None
        date_completed = contract['date_completed'] \
            if 'date_completed' in contract else None
        title = contract['title'] if 'title' in contract else None
        start_location, end_location = \
            self._identify_locations(contract, esi_client)        
        obj, created = self.update_or_create(
            handler=handler,
            contract_id=contract['contract_id'],
            defaults={
                'acceptor': acceptor,
                'acceptor_corporation': acceptor_corporation,
                'collateral': contract['collateral'],
                'date_accepted': date_accepted,
                'date_completed': date_completed,
                'date_expired': contract['date_expired'],
                'date_issued': contract['date_issued'],
                'days_to_complete': contract['days_to_complete'],
                'end_location': end_location,
                'for_corporation': contract['for_corporation'],
                'issuer_corporation': issuer_corporation,
                'issuer': issuer,                                
                'reward': contract['reward'],
                'start_location': start_location,
                'status': contract['status'],
                'title': title,
                'volume': contract['volume'],
                'pricing': None,
                'issues': None
            }                        
        )
        return obj, created

    def _identify_locations(self, contract: dict, esi_client: object) -> tuple:
        from .models import Location

        start_location, _ = Location.objects.get_or_create_from_esi(
            esi_client, contract['start_location_id']
        )
        end_location, _ = Location.objects.get_or_create_from_esi(
            esi_client, contract['end_location_id']
        )
        return start_location, end_location
    
    def _identify_contract_acceptor(self, contract: dict) -> tuple:
        from .models import EveEntity

        add_prefix = make_logger_prefix(contract['contract_id'])
        if int(contract['acceptor_id']) != 0:
            try:
                entity, _ = EveEntity.objects.get_or_create_from_esi(
                    contract['acceptor_id']
                )
                if entity.is_character:
                    try:
                        acceptor = EveCharacter.objects.get(
                            character_id=entity.id
                        )
                    except EveCharacter.DoesNotExist:
                        acceptor = EveCharacter.objects.create_character(
                            character_id=entity.id
                        )
                    try:
                        acceptor_corporation = EveCorporationInfo.objects.get(
                            corporation_id=acceptor.corporation_id
                        )
                    except EveCorporationInfo.DoesNotExist:
                        acceptor_corporation = \
                            EveCorporationInfo.objects.create_corporation(
                                corp_id=acceptor.corporation_id
                            )
                elif entity.is_corporation:
                    acceptor = None
                    try:
                        acceptor_corporation = EveCorporationInfo.objects.get(
                            corporation_id=entity.id
                        )
                    except EveCorporationInfo.DoesNotExist:
                        acceptor_corporation = \
                            EveCorporationInfo.objects.create_corporation(
                                corp_id=entity.id
                            )
                else:
                    raise ValueError(
                        'Acceptor has invalid category: {}'.format(
                            entity.category
                        )
                    )

            except Exception as ex:
                logger.exception(add_prefix(
                    'Failed to identify acceptor for this contract: {}'.format(
                        ex
                    )
                ))
                acceptor = None
                acceptor_corporation = None

        else:
            acceptor = None
            acceptor_corporation = None

        return acceptor, acceptor_corporation

    def _identify_contract_issuer(self, contract) -> tuple:
        try:
            issuer = EveCharacter.objects.get(
                character_id=contract['issuer_id']
            )
        except EveCharacter.DoesNotExist:
            issuer = EveCharacter.objects.create_character(
                character_id=contract['issuer_id']
            )

        try:
            issuer_corporation = EveCorporationInfo.objects.get(
                corporation_id=contract['issuer_corporation_id']
            )
        except EveCorporationInfo.DoesNotExist:
            issuer_corporation = EveCorporationInfo.objects.create_corporation(
                corp_id=contract['issuer_corporation_id']
            )
        return issuer_corporation, issuer

    def update_pricing(self) -> None:
        """Updates contracts with matching pricing"""
        from .models import Pricing

        def _make_key(location_id_1: int, location_id_2: int) -> str:
            return '{}x{}'.format(int(location_id_1), int(location_id_2))

        pricings = dict()
        for x in Pricing.objects.filter(is_active=True).order_by('-id'):
            pricings[_make_key(x.start_location_id, x.end_location_id)] = x
            if x.is_bidirectional:
                pricings[_make_key(x.end_location_id, x.start_location_id)] = x

        for contract in self.all():
            if (
                contract.status == self.model.STATUS_OUTSTANDING 
                or not contract.pricing
            ):
                with transaction.atomic():
                    route_key = _make_key(
                        contract.start_location_id, 
                        contract.end_location_id
                    )        
                    if route_key in pricings:
                        pricing = pricings[route_key]
                        issues_list = contract.get_price_check_issues(pricing)
                        if issues_list:
                            issues = json.dumps(issues_list)
                        else:
                            issues = None
                    else:
                        pricing = None
                        issues = None            
                        
                    contract.pricing = pricing
                    contract.issues = issues
                    contract.save()

    def send_notifications(self, force_sent=False, rate_limted=True) -> None:
        """Send notifications for outstanding contracts that have pricing"""
        add_tag = make_logger_prefix('send_notifications')
        logger.debug(add_tag('start'))

        # pilot notifications
        if FREIGHT_DISCORD_WEBHOOK_URL:
            contracts_qs = self\
                .filter(status__exact=self.model.STATUS_OUTSTANDING)\
                .exclude(pricing__exact=None)

            if not force_sent:
                contracts_qs = contracts_qs.filter(date_notified__exact=None)
            
            contracts_qs = contracts_qs.select_related()

            if contracts_qs.count() > 0:
                self._sent_pilot_notifications(
                    contracts_qs, rate_limted, add_tag
                )
        else:
            logger.debug(add_tag('FREIGHT_DISCORD_WEBHOOK_URL not configured'))

        # customer notifications        
        if FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL:
            contracts_qs = self\
                .filter(status__in=self.model.STATUS_FOR_CUSTOMER_NOTIFICATION)\
                .exclude(pricing__exact=None)
            
            contracts_qs = contracts_qs.select_related()

            if contracts_qs.count() > 0:
                self._sent_customer_notifications(
                    contracts_qs, rate_limted, force_sent, add_tag
                )
        
        else:
            logger.debug(add_tag(
                'FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL not configured'
            ))

    def _sent_pilot_notifications(self, contracts_qs, rate_limted, add_tag) -> None:
        logger.info(add_tag(
            'Trying to send pilot notifications for'
            + ' {} contracts'.format(contracts_qs.count())
        ))
        
        for contract in contracts_qs:
            if not contract.has_expired:
                contract.send_pilot_notification()
                if rate_limted:
                    sleep(1)
            else:
                logger.debug(add_tag(
                    'contract {} has expired'.format(                                    
                        contract.contract_id
                    )
                ))

    def _sent_customer_notifications(
        self, contracts_qs, rate_limted, force_sent, add_tag
    ) -> None:
        logger.debug(
            '%s: Checking %d contracts if customer '
            'notifications need to be sent', 
            add_tag(),
            contracts_qs.count()
        )
        for contract in contracts_qs:
            if contract.has_expired:
                logger.debug(
                    '%s: contract %d has expired',
                    add_tag(),
                    contract.contract_id
                )
            elif contract.has_stale_status:
                logger.debug(
                    '%s: contract %d has stale status', 
                    add_tag(),
                    contract.contract_id
                )
            else:
                contract.send_customer_notification(force_sent)
                if rate_limted:
                    sleep(1)



# Generated by Django 2.2.5 on 2019-10-28 22:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0005_auto_20191028_2155'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contracthandler',
            name='organization',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='freight.EveOrganization'),
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-10 14:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0011_contractcustomernotification'),
    ]

    operations = [
        migrations.AddField(
            model_name='contracthandler',
            name='price_per_volume_modifier',
            field=models.FloatField(blank=True, default=None, help_text='global modifier for price per volume in percent, e.g. 2.5 = +2.5%', null=True),
        ),
        migrations.AddField(
            model_name='pricing',
            name='use_price_per_volume_modifier',
            field=models.BooleanField(default=False, help_text='Whether the global price per volume modifier is used'),
        ),
    ]



from django.contrib import admin

from . import tasks
from .app_settings import FREIGHT_DEVELOPER_MODE
from .models import (
    ContractHandler, 
    Contract, 
    ContractCustomerNotification, 
    EveEntity, 
    Location, 
    Pricing
)
from .tasks import update_locations


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', '_category', '_solar_system')
    list_filter = ('category_id',)
    search_fields = ['name']
    list_select_related = True

    actions = ['update_location']

    if not FREIGHT_DEVELOPER_MODE:
        list_display_links = None
    
    def _category(self, obj):
        return obj.get_category_id_display()
    _category.admin_order_field = 'category_id'

    def _solar_system(self, obj):
        return obj.solar_system_name

    def has_add_permission(self, request):
        if FREIGHT_DEVELOPER_MODE:
            return True
        else:
            return False

    def has_change_permission(self, request):
        if FREIGHT_DEVELOPER_MODE:
            return True
        else:
            return False

    def update_location(self, request, queryset):                        
        location_ids = list()
        for obj in queryset:            
            location_ids.append(obj.pk)

        update_locations.delay(location_ids)
        self.message_user(
            request, 
            'Started updating {} locations. '
            'This can take a short while to complete.'.format(len(location_ids))
        )
    update_location.short_description = 'Update selected locations from ESI'


if FREIGHT_DEVELOPER_MODE:
    @admin.register(EveEntity)
    class EveEntityAdmin(admin.ModelAdmin):
        list_display = ('name', 'category')
        list_filter = ('category', )


@admin.register(Pricing)
class PricingAdmin(admin.ModelAdmin):
    list_display = (
        'name', 
        'start_location', 
        'end_location', 
        '_bidirectional',         
        '_default',
        '_active',
    )
    list_filter = (
        'is_bidirectional',
        'is_active',
        ('start_location', admin.RelatedOnlyFieldListFilter),
        ('end_location', admin.RelatedOnlyFieldListFilter),
    )
    list_select_related = True

    def _bidirectional(self, obj):
        return obj.is_bidirectional
    _bidirectional.boolean = True

    def _active(self, obj):
        return obj.is_active
    _active.boolean = True

    def _default(self, obj):
        return obj.is_default
    _default.boolean = True


@admin.register(ContractHandler)
class ContractHandlerAdmin(admin.ModelAdmin):
    list_display = (
        'organization', 
        'character', 
        'operation_mode', 
        'last_sync', 
        '_is_sync_ok'
    )
    actions = ('start_sync', 'send_notifications', 'update_pricing')

    if not FREIGHT_DEVELOPER_MODE:            
        readonly_fields = (
            'organization', 
            'character', 
            'operation_mode', 
            'version_hash', 
            'last_sync', 
            'last_error', 
        )

    def _is_sync_ok(self, obj):
        return obj.is_sync_ok
            
    _is_sync_ok.boolean = True
    _is_sync_ok.short_description = 'sync ok'

    def start_sync(self, request, queryset):
                        
        for obj in queryset:            
            tasks.run_contracts_sync.delay(                
                force_sync=True,
                user_pk=request.user.pk
            )            
            text = 'Started syncing contracts for: {} '.format(obj)
            text += 'You will receive a report once it is completed.'

            self.message_user(
                request, 
                text
            )
    
    start_sync.short_description = "Fetch contracts from Eve Online server"

    def send_notifications(self, request, queryset):
                        
        for obj in queryset:            
            tasks.send_contract_notifications.delay(                
                force_sent=True
            )            
            text = 'Started sending notifications for: {} '.format(obj)
            
            self.message_user(
                request, 
                text
            )
    
    send_notifications.short_description = \
        "Send notifications for outstanding contracts"

    def update_pricing(self, request, queryset):    
        del queryset
        tasks.update_contracts_pricing.delay()            
        self.message_user(
            request, 
            'Started updating pricing relations for all contracts'
        )

    update_pricing.short_description = "Update pricing info for all contracts"
    
    def has_add_permission(self, request):
        return False


@admin.register(Contract)
class ContractAdmin(admin.ModelAdmin):
    list_display = [        
        'contract_id',
        'status',
        'date_issued',
        'issuer',
        '_pilots_notified',
        '_customer_notified'
    ]
    list_filter = (
        'status',
        ('issuer', admin.RelatedOnlyFieldListFilter),
    )
    search_fields = ['issuer']

    list_select_related = True

    actions = ['send_pilots_notification', 'send_customer_notification']

    def _pilots_notified(self, contract):
        return contract.date_notified is not None

    _pilots_notified.boolean = True

    def _customer_notified(self, contract):
        return ', '.join(
            sorted([
                x.status 
                for x in contract.contractcustomernotification_set.all()
            ], reverse=True
            )
        )
    
    def send_pilots_notification(self, request, queryset):                        
        for obj in queryset:            
            obj.send_pilot_notification()
            self.message_user(
                request, 
                'Sent pilots notification for contract {} to Discord'.format(
                    obj.contract_id
                )
            )
    
    send_pilots_notification.short_description = \
        "Sent pilots notification for selected contracts to Discord"

    def send_customer_notification(self, request, queryset):                        
        for obj in queryset:            
            obj.send_customer_notification(force_sent=True)
            self.message_user(
                request, 
                'Sent customer notification for contract {} to Discord'.format(
                    obj.contract_id
                )
            )
    
    send_customer_notification.short_description = \
        "Sent customer notification for selected contracts to Discord"
    
    def has_add_permission(self, request):
        if FREIGHT_DEVELOPER_MODE:
            return True
        else:
            return False

    def has_change_permission(self, request, obj=None):
        if FREIGHT_DEVELOPER_MODE:
            return True
        else:
            return False


if FREIGHT_DEVELOPER_MODE:
    @admin.register(ContractCustomerNotification)
    class ContractCustomerNotificationAdmin(admin.ModelAdmin):
        pass



from django.conf.urls import url
from . import views

app_name = 'freight'

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(
        r'^setup_contract_handler/$', 
        views.setup_contract_handler, 
        name='setup_contract_handler'
    ),
    url(r'^add_location/$', views.add_location, name='add_location'),
    url(r'^add_location_2/$', views.add_location_2, name='add_location_2'),
    
    url(r'^calculator/$', views.calculator, name='calculator'),
    url(
        r'^calculator/(?P<pricing_pk>[0-9]+)/$', 
        views.calculator, 
        name='calculator'
    ),    
    
    url(
        r'^contract_list_active/$', 
        views.contract_list_active, 
        name='contract_list_active'
    ),
    url(
        r'^contract_list_user/$', 
        views.contract_list_user, 
        name='contract_list_user'
    ),
    url(
        r'^contract_list_data/(?P<category>.+)/$', 
        views.contract_list_data, 
        name='contract_list_data'
    ),

    url(r'^statistics/$', views.statistics, name='statistics'),
    url(
        r'^statistics_routes_data/$', 
        views.statistics_routes_data, 
        name='statistics_routes_data'
    ),
    url(
        r'^statistics_pilots_data/$', 
        views.statistics_pilots_data, 
        name='statistics_pilots_data'
    ),
    url(
        r'^statistics_pilot_corporations_data/$', 
        views.statistics_pilot_corporations_data, 
        name='statistics_pilot_corporations_data'
    ),
    url(
        r'^statistics_customer_data/$', 
        views.statistics_customer_data, 
        name='statistics_customer_data'
    ),
]



# Generated by Django 2.2.11 on 2020-04-05 16:02

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0015_auto_20200130_2328'),
    ]

    operations = [
        migrations.AddField(
            model_name='pricing',
            name='is_default',
            field=models.BooleanField(default=False, help_text='The default pricing will be preselected in the calculator. Please make sure to only mark one pricing as default.'),
        ),
        migrations.AlterField(
            model_name='location',
            name='category_id',
            field=models.PositiveIntegerField(choices=[(3, 'station'), (65, 'structure'), (0, '(unknown)')], default=0, help_text='Eve Online category ID'),
        ),
        migrations.AlterField(
            model_name='location',
            name='solar_system_id',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='Eve Online solar system ID', null=True),
        ),
        migrations.AlterField(
            model_name='location',
            name='type_id',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='Eve Online type ID', null=True),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='days_to_complete',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='Recommended days for contract completion', null=True, validators=[django.core.validators.MinValueValidator(1)]),
        ),
        migrations.AlterField(
            model_name='pricing',
            name='days_to_expire',
            field=models.PositiveIntegerField(blank=True, default=None, help_text='Recommended days for contracts to expire', null=True, validators=[django.core.validators.MinValueValidator(1)]),
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-30 20:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0008_auto_20191030_1547'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='contracthandler',
            options={'verbose_name': 'Contract Handler', 'verbose_name_plural': 'Contract Handler'},
        ),
    ]



import math

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator

from .models import Pricing


class CalculatorForm(forms.Form):             
    pricing = forms.ModelChoiceField(
        queryset=Pricing.objects
        .filter(is_active=True)
        .order_by('start_location__name', 'end_location__name'),
        label='Route',
        help_text='Pick a route for your courier contract',
        empty_label=None
    )    
    volume = forms.IntegerField(
        help_text='Est. volume of your cargo in m3',
        required=False,
        validators=[            
            MinValueValidator(0)
        ]
    )
    collateral = forms.IntegerField(
        help_text=(
            'Collaterial in ISK, must be roughly equal to the est. '
            'value of your cargo'
        ),
        required=False,
        validators=[            
            MinValueValidator(0)
        ]
    )

    def clean(self):                
        pricing = self.cleaned_data['pricing']
        issue_prefix = '⚠ Issues:'
        
        if (pricing.requires_volume() 
            and self.cleaned_data['volume'] is None
        ):
            raise ValidationError(
                '{} volume is required'.format(issue_prefix)
            )

        if (pricing.requires_collateral() 
            and self.cleaned_data['collateral'] is None
        ):
            raise ValidationError(
                '{} collateral is required'.format(issue_prefix)
            )        
        
        volume = self.cleaned_data['volume']        
        collateral = self.cleaned_data['collateral']        
        issues = pricing.get_contract_price_check_issues(
            volume,
            collateral
        )
        
        if issues:
            raise ValidationError(                
                '{} {}'.format(issue_prefix, ", ".join(issues))
            )

    def get_calculated_data(self, pricing: object) -> tuple:
        if self.is_valid():            
            if self.cleaned_data['volume']:
                volume = int(self.cleaned_data['volume'])
            else:
                volume = 0
            if self.cleaned_data['collateral']:
                collateral = int(self.cleaned_data['collateral'])        
            else:
                collateral = 0
            price = math.ceil(
                pricing.get_calculated_price(volume, collateral) / 1000000
            ) * 1000000
                
        else:
            volume = None
            collateral = None
            price = None

        return volume, collateral, price 
    

class AddLocationForm(forms.Form):
    location_id = forms.IntegerField(
        label='Location ID',
        help_text='Eve Online ID for a station or an Upwell structure'
    )    



# Generated by Django 2.2.5 on 2019-10-15 12:29

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0001_initial_new'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pricing',
            name='price_per_collateral_percent',
            field=models.FloatField(blank=True, default=None, help_text='Add-on price in % of collateral', null=True, validators=[django.core.validators.MinValueValidator(0)]),
        ),
    ]



import datetime
import logging

from django.conf import settings
from django.contrib.auth.decorators import login_required, permission_required
from django.forms import HiddenInput
from django.http import JsonResponse
from django.db.models import Count, Sum, Q
from django.shortcuts import render, redirect
from django.utils.timezone import now

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter
from esi.decorators import token_required
from esi.clients import esi_client_factory
from esi.models import Token

from . import tasks, __title__
from .app_settings import (
    FREIGHT_STATISTICS_MAX_DAYS, FREIGHT_OPERATION_MODE
)
from .models import Contract, ContractHandler, EveEntity, Location, Pricing
from .utils import (
    get_swagger_spec_path, 
    DATETIME_FORMAT, 
    messages_plus, 
    LoggerAddTag,
    create_bs_glyph_2_html
)


logger = LoggerAddTag(logging.getLogger(__name__), __package__)

ADD_LOCATION_TOKEN_TAG = 'freight_add_location_token'
CONTRACT_LIST_USER = 'user'
CONTRACT_LIST_ACTIVE = 'active'


@login_required
@permission_required('freight.basic_access')
def index(request):
    return redirect('freight:calculator')


@login_required
@permission_required('freight.view_contracts')
def contract_list_active(request):
        
    context = {
        'app_title': __title__,
        'page_title': 'Active Contracts',
        'category': CONTRACT_LIST_ACTIVE
    }        
    return render(request, 'freight/contract_list.html', context)


@login_required
@permission_required('freight.use_calculator')
def contract_list_user(request):
        
    context = {
        'app_title': __title__,
        'page_title': 'My Contracts',
        'category': CONTRACT_LIST_USER
    }        
    return render(request, 'freight/contract_list.html', context)


@login_required
@permission_required('freight.basic_access')
def contract_list_data(request, category):
    """returns list of outstanding contracts for contract_list AJAX call"""
    
    def datetime_format(x):
        return x.strftime(DATETIME_FORMAT) if x else None
    
    def character_format(x):
        return x.character_name if x else None
    
    contracts = _get_contracts_for_contract_list(category, request)

    contracts_data = list()    
    for contract in contracts:                                        
        if contract.has_pricing:            
            route_name = contract.pricing.name
            if not contract.has_pricing_errors:
                glyph = 'ok'
                color = 'green'
                tooltip_text = route_name
            else:
                glyph = 'warning-sign'
                color = 'red'                
                tooltip_text = '{}\n{}'.format(
                    route_name, 
                    '\n'.join(contract.get_issue_list())
                )
            pricing_check = create_bs_glyph_2_html(glyph, tooltip_text, color)
        else:
            route_name = ''
            pricing_check = '-'
        
        if contract.title:
            notes = create_bs_glyph_2_html('envelope', contract.title)
        else:
            notes = ''

        if settings.DEBUG:
            notes += ' {}'.format(contract.contract_id)
        
        contracts_data.append({
            'contract_id': contract.contract_id,
            'status': contract.status,
            'start_location': str(contract.start_location),
            'end_location': str(contract.end_location),
            'reward': '{:,.0f}'.format(contract.reward / 1000000),
            'collateral': '{:,.0f}'.format(contract.collateral / 1000000),
            'volume': '{:,.0f}'.format(contract.volume / 1000),
            'date_issued': datetime_format(contract.date_issued),
            'date_expired': datetime_format(contract.date_expired),
            'issuer': character_format(contract.issuer),
            'notes': notes,
            'date_accepted': datetime_format(contract.date_accepted),
            'acceptor': contract.acceptor_name,
            'has_pricing': contract.has_pricing,
            'has_pricing_errors': contract.has_pricing_errors,
            'pricing_check': pricing_check,
            'route_name': route_name,
            'is_in_progress': contract.is_in_progress,
            'is_failed': contract.is_failed,
            'is_completed': contract.is_completed,            
        })

    return JsonResponse(contracts_data, safe=False)


def _get_contracts_for_contract_list(category, request):
    if category == CONTRACT_LIST_ACTIVE:
        if not request.user.has_perm('freight.view_contracts'):
            raise RuntimeError('Insufficient permissions')
        
        else:            
            contracts = Contract.objects\
                .filter(status__in=[
                    Contract.STATUS_OUTSTANDING,
                    Contract.STATUS_IN_PROGRESS
                ])\
                .exclude(date_expired__lt=now())\
                .select_related()

    elif category == CONTRACT_LIST_USER:
        if not request.user.has_perm('freight.use_calculator'):
            raise RuntimeError('Insufficient permissions')

        else:
            user_characters = [
                x.character 
                for x in request.user.character_ownerships.select_related().all()
            ]
            contracts = Contract.objects\
                .filter(issuer__in=user_characters)\
                .filter(status__in=[
                    Contract.STATUS_OUTSTANDING,
                    Contract.STATUS_IN_PROGRESS,
                    Contract.STATUS_FINISHED,
                    Contract.STATUS_FAILED,
                ])\
                .select_related()

    else:
        raise ValueError('Invalid category: {}'.format(category))

    return contracts


@login_required
@permission_required('freight.use_calculator')
def calculator(request, pricing_pk=None):            
    from .forms import CalculatorForm
    if request.method != 'POST':
        pricing = Pricing.objects.get_or_default(pricing_pk)
        form = CalculatorForm(initial={'pricing': pricing})
        price = None        

    else:
        form = CalculatorForm(request.POST)
        request.POST._mutable = True
        pricing = Pricing.objects.get_or_default(form.data['pricing'])
        volume, collateral, price = form.get_calculated_data(pricing)         

    if pricing:
        price_per_volume_eff = pricing.price_per_volume_eff()
        if not pricing.requires_volume():
            form.fields['volume'].widget = HiddenInput()
        if not pricing.requires_collateral():
            form.fields['collateral'].widget = HiddenInput()
    else:
        price_per_volume_eff = None
    
    if price:
        if pricing.days_to_expire:
            expires_on = datetime.datetime.now(
                datetime.timezone.utc
            ) + datetime.timedelta(days=pricing.days_to_expire)
        else:
            expires_on = None
    else:
        collateral = None
        volume = None
        expires_on = None

    handler = ContractHandler.objects.first()
    if handler:
        organization_name = handler.organization.name
        availability = handler.get_availability_text_for_contracts()
    else:
        organization_name = None
        availability = None
        
    return render(
        request, 'freight/calculator.html', 
        {
            'app_title': __title__,
            'page_title': 'Reward Calculator',
            'form': form,             
            'pricing': pricing,
            'price': price,
            'organization_name': organization_name,
            'collateral': collateral if collateral is not None else 0,
            'volume': volume if volume is not None else None,
            'expires_on': expires_on,
            'availability': availability,
            'pricing_price_per_volume_eff': price_per_volume_eff
        }
    )


@login_required
@permission_required('freight.setup_contract_handler')
@token_required(scopes=ContractHandler.get_esi_scopes())
def setup_contract_handler(request, token):
    success = True
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    if ((EveEntity.get_category_for_operation_mode(
        FREIGHT_OPERATION_MODE) == EveEntity.CATEGORY_ALLIANCE)
        and token_char.alliance_id is None
    ):
        messages_plus.error(
            request, 
            'Can not setup contract handler, '
            'because {} is not a member of any alliance'.format(token_char)
        )
        success = False
    
    if success:
        try:
            owned_char = CharacterOwnership.objects.get(
                user=request.user,
                character=token_char
            )            
        except CharacterOwnership.DoesNotExist:
            messages_plus.error(
                request,
                'You can only use your main or alt characters to setup '
                + ' the contract handler. '
                + 'However, character <strong>{}</strong> is neither. '.format(
                    token_char.character_name
                )
            )
            success = False
    
    if success:
        handler = ContractHandler.objects.first()
        if handler and handler.operation_mode != FREIGHT_OPERATION_MODE:
            messages_plus.error(
                request,
                'There already is a contract handler installed for a '
                + 'different operation mode. You need to first delete the '
                + 'existing contract handler in the admin section '
                + 'before you can set up this app for a different operation mode.'
            )
            success = False

    if success:        
        organization, _ = EveEntity.objects.update_or_create_from_evecharacter(
            token_char,
            EveEntity.get_category_for_operation_mode(
                FREIGHT_OPERATION_MODE
            )
        )

        handler, _ = ContractHandler.objects.update_or_create(
            organization=organization,
            defaults={
                'character': owned_char,
                'operation_mode': FREIGHT_OPERATION_MODE
            }
        )          
        tasks.run_contracts_sync.delay(            
            force_sync=True,
            user_pk=request.user.pk
        )        
        messages_plus.success(
            request, 
            'Contract Handler setup started for '
            '<strong>{}</strong> organization '
            'with <strong>{}</strong> as sync character. '
            'Operation mode: <strong>{}</strong>. '
            'Started syncing of courier contracts. '
            'You will receive a report once it is completed.'.format(
                organization.name, 
                handler.character.character.character_name,  
                handler.operation_mode_friendly
            )
        )
        
    return redirect('freight:index')


@login_required
@token_required(scopes=Location.get_esi_scopes())
@permission_required('freight.add_location')
def add_location(request, token): 
    request.session[ADD_LOCATION_TOKEN_TAG] = token.pk
    return redirect('freight:add_location_2')


@login_required
@permission_required('freight.add_location')
def add_location_2(request): 
    from .forms import AddLocationForm
    
    if ADD_LOCATION_TOKEN_TAG not in request.session:
        raise RuntimeError('Missing token in session')
    else:
        token = Token.objects.get(pk=request.session[ADD_LOCATION_TOKEN_TAG])
    
    if request.method != 'POST':
        form = AddLocationForm()
        
    else:
        form = AddLocationForm(request.POST)
        if form.is_valid():
            location_id = form.cleaned_data['location_id']
            try:                
                client = esi_client_factory(
                    token=token, 
                    spec_file=get_swagger_spec_path()
                )
            
                location, created = Location.objects.update_or_create_from_esi(
                    client, 
                    location_id,
                    add_unknown=False
                )
                action_txt = 'Added:' if created else 'Updated:'
                messages_plus.success(
                    request,
                    '{} <strong>{}</strong>'.format(
                        action_txt,                        
                        location.name
                    )
                )
                return redirect('freight:add_location_2')    

            except Exception as ex:
                messages_plus.error(
                    request,                    
                    'Failed to add location with token from {} '
                    'for location ID {}: {}'.format(
                        token.character_name, location_id, type(ex).__name__
                    )
                )
        
    return render(
        request, 'freight/add_location.html', 
        {            
            'app_title': __title__,
            'page_title': 'Add / Update Location',
            'form': form,
            'token_char_name': token.character_name
        }
    )


@login_required
@permission_required('freight.view_statistics')
def statistics(request):

    context = {
        'app_title': __title__,
        'page_title': 'Statistics',
        'max_days': FREIGHT_STATISTICS_MAX_DAYS
    }        
    return render(request, 'freight/statistics.html', context)


@login_required
@permission_required('freight.view_statistics')
def statistics_routes_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contract__status__exact=Contract.STATUS_FINISHED) \
        & Q(contract__date_completed__gte=cutoff_date)
    route_totals = Pricing.objects.select_related() \
        .annotate(contracts_count=Count(
            'contract', filter=finished_contracts
        )) \
        .annotate(rewards=Sum(
            'contract__reward', filter=finished_contracts
        )) \
        .annotate(collaterals=Sum(
            'contract__collateral', filter=finished_contracts
        )) \
        .annotate(pilots=Count(
            'contract__acceptor', distinct=True, filter=finished_contracts
        )) \
        .annotate(
            customers=Count(
                'contract__issuer', 
                distinct=True, 
                filter=finished_contracts
            )
    )   

    totals = list()
    for route in route_totals:
        if route.contracts_count > 0:
            if route.rewards:
                rewards = route.rewards / 1000000
            else:
                rewards = 0

            if route.collaterals:
                collaterals = route.collaterals / 1000000
            else:
                collaterals = 0
            
            totals.append({
                'name': route.name,
                'contracts': '{:,}'.format(route.contracts_count),
                'rewards': '{:,.0f}'.format(rewards),
                'collaterals': '{:,.0f}'.format(collaterals),
                'pilots': '{:,}'.format(route.pilots),
                'customers': '{:,}'.format(route.customers),                
            })

    return JsonResponse(totals, safe=False)


@login_required
@permission_required('freight.view_statistics')
def statistics_pilots_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(FREIGHT_STATISTICS_MAX_DAYS)
    
    finished_contracts = \
        Q(contract_acceptor__status__exact=Contract.STATUS_FINISHED) \
        & Q(contract_acceptor__date_completed__gte=cutoff_date)
    
    pilot_totals = EveCharacter.objects\
        .exclude(contract_acceptor__exact=None)\
        .select_related() \
        .annotate(
            contracts_count=Count(
                'contract_acceptor', filter=finished_contracts
            )) \
        .annotate(rewards=Sum(
            'contract_acceptor__reward', filter=finished_contracts
        )) \
        .annotate(collaterals=Sum(
            'contract_acceptor__collateral', filter=finished_contracts
        ))        

    totals = list()
    for pilot in pilot_totals:
        if pilot.contracts_count > 0:
            if pilot.rewards:
                rewards = pilot.rewards / 1000000
            else:
                rewards = 0

            if pilot.collaterals:
                collaterals = pilot.collaterals / 1000000
            else:
                collaterals = 0
            
            totals.append({
                'name': pilot.character_name,
                'corporation': pilot.corporation_name,
                'contracts': '{:,}'.format(pilot.contracts_count),
                'rewards': '{:,.0f}'.format(rewards),
                'collaterals': '{:,.0f}'.format(collaterals),            
            })

    return JsonResponse(totals, safe=False)


@login_required
@permission_required('freight.view_statistics')
def statistics_pilot_corporations_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(FREIGHT_STATISTICS_MAX_DAYS)
    
    finished_contracts = \
        Q(
            contract_acceptor_corporation__status__exact=Contract.STATUS_FINISHED
        ) & Q(
            contract_acceptor_corporation__date_completed__gte=cutoff_date
        )
    
    corporation_totals = EveCorporationInfo.objects\
        .exclude(contract_acceptor_corporation__exact=None)\
        .select_related() \
        .annotate(contracts_count=Count(
            'contract_acceptor_corporation', 
            filter=finished_contracts
        )) \
        .annotate(rewards=Sum(
            'contract_acceptor_corporation__reward', filter=finished_contracts
        )) \
        .annotate(collaterals=Sum(
            'contract_acceptor_corporation__collateral', filter=finished_contracts
        ))        

    totals = list()
    for corporation in corporation_totals:
        if corporation.contracts_count > 0:
            if corporation.rewards:
                rewards = corporation.rewards / 1000000
            else:
                rewards = 0

            if corporation.collaterals:
                collaterals = corporation.collaterals / 1000000
            else:
                collaterals = 0
                        
            totals.append({
                'name': corporation.corporation_name,
                'alliance': corporation.alliance.alliance_name
                if corporation.alliance else '',
                'contracts': '{:,}'.format(corporation.contracts_count),
                'rewards': '{:,.0f}'.format(rewards),
                'collaterals': '{:,.0f}'.format(collaterals),            
            })

    return JsonResponse(totals, safe=False)


@login_required
@permission_required('freight.view_statistics')
def statistics_customer_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contract_issuer__status__exact=Contract.STATUS_FINISHED) \
        & Q(contract_issuer__date_completed__gte=cutoff_date)
    customer_totals = EveCharacter.objects\
        .exclude(contract_issuer__exact=None)\
        .select_related() \
        .annotate(contracts_count=Count(
            'contract_issuer', filter=finished_contracts
        )) \
        .annotate(rewards=Sum(
            'contract_issuer__reward', filter=finished_contracts
        )) \
        .annotate(collaterals=Sum(
            'contract_issuer__collateral', filter=finished_contracts
        ))

    totals = list()
    for customer in customer_totals:        
        if customer.contracts_count > 0:
            if customer.rewards:
                rewards = customer.rewards / 1000000
            else:
                rewards = 0

            if customer.collaterals:
                collaterals = customer.collaterals / 1000000
            else:
                collaterals = 0
            
            totals.append({
                'name': customer.character_name,
                'corporation': customer.corporation_name,
                'contracts': '{:,}'.format(customer.contracts_count),
                'rewards': '{:,.0f}'.format(rewards),
                'collaterals': '{:,.0f}'.format(collaterals),            
            })

    return JsonResponse(totals, safe=False)



from django.conf import settings
from .utils import clean_setting

# Sets the number minutes until a delayed sync will be recognized as error
FREIGHT_CONTRACT_SYNC_GRACE_MINUTES = \
    clean_setting('FREIGHT_CONTRACT_SYNC_GRACE_MINUTES', 30)


# Enables features for developers, e.g. write access to all models in admin
FREIGHT_DEVELOPER_MODE = \
    clean_setting('FREIGHT_DEVELOPER_MODE', False)


# Webhook URL used for notifications if defined
FREIGHT_DISCORD_WEBHOOK_URL = \
    clean_setting('FREIGHT_DISCORD_WEBHOOK_URL', None, required_type=str)


# Will be shown as "user name" instead of "Alliance Freight" 
# for notifications if defined
FREIGHT_DISCORD_AVATAR_NAME = \
    clean_setting('FREIGHT_DISCORD_AVATAR_NAME', None, required_type=str)


# when set true will no longer set name and avatar for webhooks
FREIGHT_DISCORD_DISABLE_BRANDING = \
    clean_setting('FREIGHT_DISCORD_DISABLE_BRANDING', False)


# Mentions to appear upfront to any notification
# Typical values are: None, '@here' and '@everyone'.
FREIGHT_DISCORD_MENTIONS = \
    clean_setting('FREIGHT_DISCORD_MENTIONS', None, required_type=str)


# max days back considered when calculating statistics
FREIGHT_STATISTICS_MAX_DAYS = \
    clean_setting('FREIGHT_STATISTICS_MAX_DAYS', 90, min_value=1)

# Webhook URL used for notifications to customers if defined
FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL = \
    clean_setting('FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL', None, required_type=str)


# defines after how many hours a status becomes stale
# a stale status will not be reported to customers
FREIGHT_HOURS_UNTIL_STALE_STATUS = \
    clean_setting('FREIGHT_HOURS_UNTIL_STALE_STATUS', 24)


# Whether to show full location names in the route dropdown of the calculator
FREIGHT_FULL_ROUTE_NAMES = \
    clean_setting('FREIGHT_FULL_ROUTE_NAMES', False)


# modes of operation for Alliance Freight
FREIGHT_OPERATION_MODE_MY_ALLIANCE = 'my_alliance'
FREIGHT_OPERATION_MODE_MY_CORPORATION = 'my_corporation'
FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE = 'corp_in_alliance'
FREIGHT_OPERATION_MODE_CORP_PUBLIC = 'corp_public'

FREIGHT_OPERATION_MODES = [
    (FREIGHT_OPERATION_MODE_MY_ALLIANCE, 'My Alliance'),
    (FREIGHT_OPERATION_MODE_MY_CORPORATION, 'My Corporation'),
    (FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE, 'Corporation in my Alliance'),
    (FREIGHT_OPERATION_MODE_CORP_PUBLIC, 'Corporation public'),
]

if (hasattr(settings, 'FREIGHT_OPERATION_MODE') 
    and settings.FREIGHT_OPERATION_MODE in [x[0] for x in FREIGHT_OPERATION_MODES]
):
    FREIGHT_OPERATION_MODE = settings.FREIGHT_OPERATION_MODE
else:
    FREIGHT_OPERATION_MODE = FREIGHT_OPERATION_MODE_MY_ALLIANCE



# Generated by Django 2.2.5 on 2019-11-27 15:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0010_auto_20191108_2220'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContractCustomerNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('outstanding', 'outstanding'), ('in_progress', 'in progress'), ('finished_issuer', 'finished issuer'), ('finished_contractor', 'finished contractor'), ('finished', 'finished'), ('canceled', 'canceled'), ('rejected', 'rejected'), ('failed', 'failed'), ('deleted', 'deleted'), ('reversed', 'reversed')], max_length=32)),
                ('date_notified', models.DateTimeField(help_text='datetime of notification')),
                ('contract', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='freight.Contract')),
            ],
            options={
                'unique_together': {('contract', 'status')},
            },
        ),
    ]



# Generated by Django 2.2.9 on 2020-01-30 23:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0014_auto_20191214_1712'),
    ]

    operations = [
        migrations.RenameField(
            model_name='pricing',
            old_name='active',
            new_name='is_active',
        ),
        migrations.AddField(
            model_name='pricing',
            name='is_bidirectional',
            field=models.BooleanField(default=True, help_text='Whether this pricing is valid for contracts in either direction or only the one specified'),
        ),
    ]



# Generated by Django 2.2.5 on 2019-11-08 22:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0009_auto_20191030_2046'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contracthandler',
            name='operation_mode',
            field=models.CharField(default='my_alliance', help_text='defines what kind of contracts are synced', max_length=32),
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-14 15:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0010_alliance_ticker'),
        ('freight', '0012_auto_20191210_1406'),
    ]

    operations = [
        migrations.AddField(
            model_name='contract',
            name='acceptor_corporation',
            field=models.ForeignKey(blank=True, default=None, help_text='corporation of acceptor', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='contract_acceptor_corporation', to='eveonline.EveCorporationInfo'),
        ),
        migrations.AlterField(
            model_name='contract',
            name='acceptor',
            field=models.ForeignKey(blank=True, default=None, help_text='character of acceptor or None if accepted by corp', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='contract_acceptor', to='eveonline.EveCharacter'),
        ),
        migrations.AlterField(
            model_name='contract',
            name='issuer_corporation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contract_issuer_corporation', to='eveonline.EveCorporationInfo'),
        ),
        migrations.AlterField(
            model_name='eveorganization',
            name='category',
            field=models.CharField(choices=[('alliance', 'Alliance'), ('corporation', 'Corporation'), ('character', 'Character')], max_length=32),
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-29 18:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('freight', '0006_auto_20191028_2204'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='contracthandler',
            options={'verbose_name': 'Contract Handler [my_alliance]', 'verbose_name_plural': 'Contract Handler [my_alliance]'},
        ),
        migrations.AddField(
            model_name='contracthandler',
            name='operation_mode',
            field=models.CharField(choices=[('my_alliance', 'My Alliance'), ('my_corporation', 'My Corporation')], default='my_alliance', help_text='defines what kind of contracts are synced', max_length=32),
        ),
    ]



from django.apps import AppConfig
from . import __version__


class FreightConfig(AppConfig):
    name = 'freight'
    label = 'freight'
    verbose_name = 'Freight v{}'.format(__version__)

    def ready(self):
        from . import signals   # noqa F401
